/* z_pq.c (priority queues)

   Ambros Marzetta

   94-10-24

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 17:14:52 +0100 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.2  1995/08/02  13:28:46  marzetta
 * Portability to macintosh
 *
 */

#include <stdlib.h> /* malloc */
#include <stdio.h> /* sprintf for z_pq_InfoString */
#include <string.h> /* memcpy */
#include <assert.h>
#define _Z_PQ_C_
#include "s_pq.h"

/* Types ----------------------------------------------------------- */

struct z_pqT {
  z_pq_CompProcP compareProc;
  long numElem;			/* number of elements in queue */
  long numIns, numDel, numDelAbove, numMin, maxElem; /* for statistics */
  z_pq_typeT type;
};

/* Heap implementation ------------------------------------------------ */

/* A heap in virtual memory is maybe not the most efficient implementation. */

#define HEAPSIZEINCREMENT 100

/* z_pq_HeapP	<-> z_pqP */
/* refHeapP	<-> z_pq_refP */

typedef struct z_pq_HeapT {
  struct z_pqT common;
  int heapSize;
  z_pq_elemP *heap;
} *z_pq_HeapP;

typedef z_pq_elemP *refHeapP;

static z_pq_HeapP NewPriorityQueueHeap(void)
{
  /* Allocate an empty priority queue. */
  z_pq_HeapP myPQ;

  myPQ = (z_pq_HeapP) malloc(sizeof(struct z_pq_HeapT));
  if (myPQ) {
    myPQ->heapSize = HEAPSIZEINCREMENT;
    myPQ->heap = (z_pq_elemP*) malloc(sizeof(z_pq_elemP) * HEAPSIZEINCREMENT);
  }
#ifndef RETURNNULL
  assert(myPQ);
  assert(myPQ->heap);
#endif
  return myPQ;
}

static void FreePriorityQueueHeap(z_pq_HeapP pq)
{
  assert(pq);
  assert(pq->heap);
  free(pq->heap);
  free(pq);
}

static z_pq_elemP GetObjectHeap(z_pqP pq, refHeapP where)
{
  /* Retrieve the object referenced by where. */
  return *where;
}

static void DoInsert(z_pq_elemP *heap, int *i, z_pq_elemP x,
	      z_pq_CompProcP compare)
{
  /* Insert x at position i and move towards the root */
  for (;;) {
    if (*i <= 1) break;
    if (compare(x, heap[*i/2]) >= 0) break;
    heap[*i] = heap[*i/2];
    *i = *i / 2;
  }
  heap[*i] = x;
}

static refHeapP InsertHeap(z_pq_HeapP pq, z_pq_elemP x)
{
  /* Insert x into the priority queue. */
  int i;
  
  pq->common.numElem ++;
  if (pq->common.numElem == pq->heapSize) {
    z_pq_elemP *newheap;
    int newheapsize;

    newheapsize = pq->heapSize + HEAPSIZEINCREMENT;
    newheap = (z_pq_elemP*) malloc(sizeof(z_pq_elemP) * newheapsize);
#ifndef RETURNNULL
    assert(newheap);
#endif
    /*bcopy(pq->heap, newheap, sizeof(z_pq_elemP) * pq->heapSize);*/
    memcpy(newheap, pq->heap, sizeof(z_pq_elemP) * pq->heapSize);
    pq->heapSize = newheapsize;
    free(pq->heap);
    pq->heap = newheap;
  }
  i = pq->common.numElem;
  DoInsert(pq->heap, &i, x, pq->common.compareProc);
  return pq->heap + i;
} /* InsertHeap */

static refHeapP MinimumHeap(z_pq_HeapP pq)
{
  /* Return a reference to the minimum without deleting it. */
  assert(pq->common.numElem >= 1);
  return pq->heap + 1;		/* The first element has index 1 */
}

static void Restore(z_pq_elemP *heap, int l, int r, z_pq_CompProcP compare)
{
  /* see Nievergelt / Hinrichs, p. 231 */
  int i, j;
  z_pq_elemP h;
  
  i = l;
  while (i <= r/2) {
    if (2*i < r)
      if (compare(heap[2*i+1], heap[2*i]) < 0)
	j = 2*i+1;
      else
	j = 2*i;
    else
      j = 2*i;
    if (compare(heap[j], heap[i]) < 0) {
      h = heap[j];
      heap[j] = heap[i];
      heap[i] = h;
      i = j;
    } else
      i = r;
  }
}

static void DeleteHeap(z_pq_HeapP pq, refHeapP ref)
{
  /* Delete the element referenced by ref. */
  int i;
  
  i = ref - pq->heap;
  assert(i > 0 && i <= pq->common.numElem);
  pq->heap[i] = pq->heap[pq->common.numElem];
  pq->common.numElem --;
  if (i == 1 || pq->common.compareProc(pq->heap[i], pq->heap[i/2]) > 0)
    Restore(pq->heap, i, pq->common.numElem, pq->common.compareProc);
  else
    DoInsert(pq->heap, &i, pq->heap[i], pq->common.compareProc);
}

static void DeleteAboveHeap(z_pq_HeapP pq, z_pq_elemP upperBound,
			    whatToDoProcP whatToDo)
{
  z_pq_elemP *heap;
  int l, r, ll;
  
  heap = pq->heap;
  l = 1;
  r = pq->common.numElem;
  for (;;) {
    /* Invariant: heap[1..l-1] < upperBound
                  heap[l..r] need yet to be compared with upperBound
		  heap[1..r] has the heap property
		  */
    while (l <= r && pq->common.compareProc(heap[l], upperBound) < 0)
      l++;
    while (l <= r && pq->common.compareProc(heap[r], upperBound) >= 0) {
      if (whatToDo) whatToDo(heap[r]);
      pq->common.numDelAbove ++;
      r--;
    }
    if (l > r) break;
    /* heap[l] >= upperBound, heap[r] < upperBound, l < r */
    if (whatToDo) whatToDo(heap[l]);
    pq->common.numDelAbove ++;
    ll = l;
    DoInsert(heap, &ll, heap[r], pq->common.compareProc);
    l++;
    r--;
  }
  pq->common.numElem = r;
}

static void ForAllHeap(z_pq_HeapP pq, whatToDoProcP whatToDo)
{
  int i;

  for (i=1; i<=pq->common.numElem; i++)
    whatToDo(pq->heap[i]);
}

/* Implementation table -------------------------------------------- */

typedef z_pqP (*npqProc)(void);
static z_pqP (*NewPriorityQueue[z_pq_numtypes])(void) =
{0, (npqProc)NewPriorityQueueHeap};

typedef void (*fpqProc)(z_pqP pq);
static void (*FreePriorityQueue[z_pq_numtypes])(z_pqP pq) =
{0, (fpqProc)FreePriorityQueueHeap};

static z_pq_elemP (*GetObject[z_pq_numtypes])(z_pqP, z_pq_refP) =
{0, (z_pq_elemP (*)(z_pqP, z_pq_refP))GetObjectHeap};

static z_pq_refP (*Insert[z_pq_numtypes])(z_pqP pq, z_pq_elemP x) =
{0, (z_pq_refP (*)(z_pqP, z_pq_elemP))InsertHeap};

static z_pq_refP (*Minimum[z_pq_numtypes])(z_pqP pq) =
{0, (z_pq_refP (*)(z_pqP pq))MinimumHeap};

static void (*Delete[z_pq_numtypes])(z_pqP pq, z_pq_refP ref) =
{0, (void (*)(z_pqP pq, z_pq_refP ref))DeleteHeap};

static void (*DeleteAbove[z_pq_numtypes])(z_pqP, z_pq_elemP, whatToDoProcP) =
{0, (void (*)(z_pqP, z_pq_elemP, whatToDoProcP))DeleteAboveHeap};

static void (*ForAll[z_pq_numtypes])(z_pqP pq, whatToDoProcP whatToDo) =
{0, (void (*)(z_pqP pq, whatToDoProcP whatToDo))ForAllHeap};

/* Procedures (general) --------------------------------------------- */

z_pqP z_pq_NewPriorityQueue(z_pq_typeT type, z_pq_CompProcP compare)
{
  /* Allocate an empty priority queue. Choose an implementation
     of type type. */
  z_pqP myQueue;

  assert(type < z_pq_numtypes);
  assert(NewPriorityQueue[type]);
  assert(compare);
  myQueue = NewPriorityQueue[type]();
  myQueue->compareProc = compare;
  myQueue->numElem = 0;
  myQueue->numIns = 0;
  myQueue->numDel = 0;
  myQueue->numDelAbove = 0;
  myQueue->numMin = 0;
  myQueue->maxElem = 0;
  myQueue->type = type;
  return myQueue;
}

void z_pq_FreePriorityQueue(z_pqP pq)
{
  /* Deallocate the priority queue pq, which must be empty. */
  FreePriorityQueue[pq->type](pq);
}

z_pq_elemP z_pq_GetObject(z_pqP pq, z_pq_refP where)
{
  /* Retrieve the object referenced by where. */
  return GetObject[pq->type](pq, where);
}

z_pq_refP z_pq_Insert(z_pqP pq, z_pq_elemP x)
{
  /* Insert x into the priority queue. */
  pq->numIns ++;
  if (pq->numElem+1 > pq->maxElem) pq->maxElem = pq->numElem+1;
  return Insert[pq->type](pq, x); /* must update numElem */
}

z_pq_refP z_pq_Minimum(z_pqP pq)
{
  /* Return a reference to the minimum without deleting it. */
  pq->numMin ++;
  return Minimum[pq->type](pq);
}

int z_pq_isEmpty(z_pqP pq)
{
  /* Test whether the queue is empty. */
  return pq->numElem == 0;
}

long z_pq_numElements(z_pqP pq)
{
  /* Return the number of elements in pq. */
  return pq->numElem;
}

void z_pq_Delete(z_pqP pq, z_pq_refP ref)
{
  /* Delete the element referenced by ref. */
  pq->numDel ++;
  Delete[pq->type](pq, ref);	/* must decrement numElem */
}

void z_pq_DeleteAbove(z_pqP pq, z_pq_elemP upperBound, whatToDoProcP whatToDo)
{
  /* Remove all elements >= upperBound from the priority queue. Call
     whatToDo for each of them. WhatToDo must not access the priority
     queue. */
  DeleteAbove[pq->type](pq, upperBound, whatToDo);
}

void z_pq_ForAll(z_pqP pq, whatToDoProcP whatToDo)
{
  /* Call whatToDo for all elements of the priority queue in an unspecified
     order. WhatToDo must not access the priority queue. */
  ForAll[pq->type](pq, whatToDo);
}

char *z_pq_InfoString(z_pqP pq)
{
  /* Return statistical information about the queue. */
  static char s[80];

  sprintf(s, "elem = %ld, max size = %ld; "
	  "ins = %ld, del = %ld, delab = %ld, min = %ld",
	  pq->numElem, pq->maxElem, pq->numIns, pq->numDel, pq->numDelAbove,
	  pq->numMin);
  return s;
}
